/*
 *
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/mm.h>
#include <asm/io.h>     /* for virt_to_phys */
#include "pvrusb2-hdw.h"
#include "pvrusb2.h"
#include "pvrusb2-buffer.h"

/*
 * pvr_buff_complete()
 *
 * This is called after an urb used for video stream is back.
 *
 */

static void pvr_buffer_complete (struct urb *purb, struct pt_regs *regs)
{
	struct pvr_buffer *b = (struct pvr_buffer *) purb->context;
	struct pvr2_v4l2_stream *stream = b->stream;
    unsigned long flags;
//    u8 *t = purb->transfer_buffer;

//    pvr_debug("pvr_buffer_complete %i - %02x %02x %02x %02x %02x %02x ",
//        purb->actual_length, t[0],t[1],t[2],t[3],t[4],t[5] );

	if (purb->status != -ENOENT) {
        /* urb was not canceled. Put it in our rx ring. */
        b->data = purb->transfer_buffer;
        b->size = purb->actual_length;
        spin_lock_irqsave(&stream->ring_lock, flags);
        stream->rx_buffer[stream->rx2 % PVR_NBUFF] = b;
        stream->rx2++;
        spin_unlock_irqrestore(&stream->ring_lock, flags);
        if (waitqueue_active(&stream->wait_data))
		    wake_up(&stream->wait_data);
    }
    if (atomic_dec_and_test(&stream->submitted_urbs)){
        pvr_debug("pvr_buffer_complete : no more urbs submitted !");
        wake_up(&stream->urbs_collected);
    }
    b->submitted = 0;
}

int pvr_buffer_alloc(struct pvr2_v4l2_stream *stream)
{
    int i, pipe;
    struct pvr_buffer *b;
    struct pvr2_hdw *hdw = stream->v4lp->channel.hdw;

    pvr_debug("pvr_alloc_buffer");

    if (stream->mem_state != _freed){
        err("illegal pvr_alloc_buffer");
        return -EBUSY;
    }

    memset(stream->buffer, 0, sizeof(struct pvr_buffer)*PVR_NBUFF);

    if (stream->buffer_count > PVR_NBUFF){
        err("to many video buffers!");
        return -ENOMEM;
    }

    stream->mem_state = _allocated;

    for (i = 0; i < stream->buffer_count; i++){
        b = &stream->buffer[i];
        b->purb = usb_alloc_urb(0, GFP_KERNEL);
        if (b->purb == NULL)
            goto err;
        /* remember, we want page aligned buffers for mmap */
        b->buff = kmalloc(PVR_VID_BUFFSIZE + PAGE_SIZE - 1, GFP_KERNEL);
        if (b->buff == NULL)
            goto err;
        memset(b->buff, 0, PVR_VID_BUFFSIZE + PAGE_SIZE - 1);
        b->stream = stream;
        pipe = usb_rcvbulkpipe(pvr2_hdw_get_dev(hdw),stream->endpoint);
        usb_fill_bulk_urb (b->purb,
			   pvr2_hdw_get_dev(hdw),
			   pipe,
                           (void *) PAGE_ALIGN((unsigned long)b->buff),
                           PVR_VID_BUFFSIZE, pvr_buffer_complete, b);
      	b->purb->transfer_flags = 0;
        b->submitted = 0;
        b->map = NULL;
        b->index = i;
    }

    return 0;

err:
    pvr_buffer_free(stream);

    return -ENOMEM;
}

/* Safe even if nothing allocated */

void pvr_buffer_free(struct pvr2_v4l2_stream *stream)
{
    int i;
    struct pvr_buffer *b;

    pvr_debug("pvr_free_buffer");
    for (i = 0; i < PVR_NBUFF; i++){
        b = &stream->buffer[i];
        if (b->purb){
            usb_free_urb(b->purb);
            b->purb = NULL;
            if (b->buff)
                kfree(b->buff);
        }
    }
    stream->mem_state = _freed;
}

/* Call with lock held */

void pvr_buffer_unlink(struct pvr2_v4l2_stream *stream)
{
    int i;

    pvr_debug("pvr_unlink_all_buffer");

    for(i = 0 ; i < stream->buffer_count; i++)
        if (stream->buffer[i].submitted) {
            usb_kill_urb(stream->buffer[i].purb);
            stream->buffer[i].submitted = 0;
        }
    /* Is there a small chance that an urb is back, but his completion
     * handler not yet complete ?
     * We must wait a bit to be sure we have everybody...
     */

    if (wait_event_interruptible_timeout(stream->urbs_collected,
                                         atomic_read(&stream->submitted_urbs)
                                          == 0, HZ) == 0)
        err("timeout waiting for urbs !");

    /* reinitialize the rx ring. */
    stream->rx1 = 0;
    stream->rx2 = 0;
}

int pvr_buffer_submit(struct pvr2_v4l2_stream *stream)
{
    int i, ret = 0;
    struct pvr2_hdw *hdw = stream->v4lp->channel.hdw;

    pvr_debug("pvr_submit_all_buffer");

    if (!pvr2_hdw_dev_ok(hdw)) {
        return -EIO;
    }

    for (i = 0 ; i < stream->buffer_count ; i++) {
        stream->buffer[i].submitted = 1;
        atomic_inc(&stream->submitted_urbs);
        ret = usb_submit_urb(stream->buffer[i].purb, GFP_KERNEL);
	if (ret != 0) break;
	stream->buffer[i].submitted = 1;
    }

    if (ret)
        pvr_buffer_unlink(stream);

    return ret;
}

int pvr_buffer_submit_one(struct pvr2_v4l2_stream *stream, int i)
{
    int ret;
    struct pvr2_hdw *hdw = stream->v4lp->channel.hdw;

    if (stream->buffer[i].submitted ||
	!pvr2_hdw_dev_ok(hdw)) {
        return -EINVAL;
    }

    ret = usb_submit_urb(stream->buffer[i].purb, GFP_KERNEL);
    if (ret == 0) {
        atomic_inc(&stream->submitted_urbs);
        stream->buffer[i].submitted = 1;
    }

    return ret;
}

static void pvr_vm_open (struct vm_area_struct *vma)
{
	struct pvr_mapping *map = vma->vm_private_data;

    pvr_debug("pvr_vm_open");

	map->count++;
}

static void pvr_vm_close(struct vm_area_struct *vma)
{
    int i;
	struct pvr_mapping *map = vma->vm_private_data;

    pvr_debug("pvr_vm_close");

    map->count--;

    if (map->count == 0) {
        struct pvr2_v4l2_stream *stream = map->stream;

        pvr_debug("pvr_vm_close : unmapping memory");

        for (i = 0 ; i < stream->buffer_count ; i++) {
            if (stream->buffer[i].map == map)
                stream->buffer[i].map = NULL;

        }
        kfree(map);
        down(&stream->stream_sem); do {
	    stream->map_count--;
	    if (stream->map_count == 0)
		stream->mem_state = _allocated;
        } while (0); up(&stream->stream_sem);
    }
}

static struct page*
pvr_vm_nopage(struct vm_area_struct *vma, unsigned long vaddr,
		   int *type)
{
    struct page *page;
    unsigned long buff;
    unsigned long offset = vaddr - vma->vm_start + (vma->vm_pgoff << PAGE_SHIFT);
    struct pvr_mapping *map = vma->vm_private_data;

    pvr_debug("pvr_vm_nopage");

    buff = (unsigned long) (map->stream->buffer[offset / PVR_VID_BUFFSIZE].buff);
    page = virt_to_page(PAGE_ALIGN(buff) + (offset % PVR_VID_BUFFSIZE));
    if (!page)
 		return NOPAGE_OOM;
    SetPageReserved(page);
    get_page(page);
    if (type)
        *type = VM_FAULT_MINOR;

    return page;
}

static struct vm_operations_struct pvr_vm_ops = {
	.open = pvr_vm_open,
	.close = pvr_vm_close,
    .nopage = pvr_vm_nopage
};

/* For V4L1 compat, we need to be able to map all buffers in one shot. */

int pvr_map_buffer(struct pvr2_v4l2_stream *stream, struct vm_area_struct *vma)
{
    int i, first, last;
    unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
    unsigned long size = vma->vm_end - vma->vm_start;
    struct pvr_mapping *map;

    pvr_debug("pvr_map_buffer");

    if (stream->mem_state < _allocated) {
        err("illegal pvr_map_buffer");
        return -EINVAL;
    }

    if (size == 0 || size % PVR_VID_BUFFSIZE != 0){
        err("pvr_map_buffer : wrong size !");
        return -EINVAL;
    }

    /* find first and last buffer to remap */
    first = offset / PVR_VID_BUFFSIZE;
    last = first + size / PVR_VID_BUFFSIZE - 1;

    /* Verify none of them is already mapped. */
    for (i = first ; i <= last ; i++)
        if (stream->buffer[i].map){
            err("buffer %i already mapped", i);
            return -EBUSY;
        }
    /* map'em all */
    map = kmalloc(sizeof(struct pvr_mapping), GFP_KERNEL);
    if (map == NULL)
        return -ENOMEM;
    map->stream = stream;
    map->count = 1;
    vma->vm_private_data = map;
    vma->vm_ops = &pvr_vm_ops;
    vma->vm_flags |= VM_RESERVED | VM_SHARED;
    stream->map_count++;
    stream->mem_state = _mapped;

    for (i = first ; i <= last ; i++)
        stream->buffer[i].map = map;

    return 0;
}

/*
 * Wait for buffer(s) with at least 'size' bytes to be available. Timeout of 0
 * means non blocking io.
 * Returns first buffer index on success,  < 0 error code on failure.
 * Detects stream break.
 */

int pvr_buffer_wait(struct pvr2_v4l2_stream *stream, unsigned long timeout)
{
    if (timeout) {
        if (wait_event_interruptible_timeout(stream->wait_data,
                                             stream->rx2 - stream->rx1 != 0,
                                             timeout) > 0) {
            stream->last_buffer_timestamp = jiffies;
            return stream->rx1 % PVR_NBUFF;
        } else
            return -EIO;
    } else {
        if (stream->rx2 - stream->rx1){
            stream->last_buffer_timestamp = jiffies;
            return stream->rx1 % PVR_NBUFF;
        } else {
            if (time_after(jiffies, stream->last_buffer_timestamp + 2 * HZ))
                /* broken stream. Now it's time to kiss your ass goodbye... */
                return -EIO;
            else
                return -EAGAIN;
        }
    }
}

/* Check if enough data (n bytes) is queued in rx_ring. */

int pvr_buffer_in(struct pvr2_v4l2_stream *stream, int n)
{
    int i = stream->rx1;

    while ((i != stream->rx2) && (n > 0))
        n -= stream->rx_buffer[i++ % PVR_NBUFF]->size;
    return (n > 0) ? (time_after(jiffies, stream->last_buffer_timestamp + 2 * HZ) ?
               -EIO : -EAGAIN)
            : 0;
}

void pvr_buffer_dq(struct pvr2_v4l2_stream *stream)
{
    stream->rx_buffer[stream->rx1 % PVR_NBUFF]->data = NULL;
    stream->rx1++;
}

