/*
 *  $Id: pvrusb2-io.c,v 1.4 2005/04/09 04:45:04 isely Exp $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2-io.h"
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <asm/semaphore.h>

struct pvr2_stream {
    /* Buffers queued for reading */
    struct list_head queued_list;
    unsigned int q_count;
    /* Buffers with retrieved data */
    struct list_head ready_list;
    unsigned int r_count;
    /* Buffers available for use */
    struct list_head idle_list;
    unsigned int i_count;
    /* Pointers to all buffers */
    struct pvr2_buffer **buffers;
    /* Array size of buffers */
    unsigned int buffer_slot_count;
    /* Total buffers actually in circulation */
    unsigned int buffer_total_count;
    /* Designed number of buffers to be in circulation */
    unsigned int buffer_target_count;
    /* Executed when ready list become non-empty */
    pvr2_stream_callback callback_func;
    void *callback_data;
    /* Context for transfer pipe */
    struct usb_device *dev;
    unsigned int pipe;
    /* Overhead for mutex enforcement */
    spinlock_t list_lock;
    struct semaphore sem;
};

struct pvr2_buffer {
    int id;
    enum pvr2_buffer_state state;
    void *ptr;               /* Pointer to storage area */
    unsigned int max_count;  /* Size of storage area */
    unsigned int used_count; /* Amount of valid data in storage area */
    int status;              /* Transfer result status */
    struct pvr2_stream *stream;
    struct list_head list_overhead;
    struct urb *purb;
};

static void pvr2_buffer_remove(struct pvr2_buffer *bp)
{
    unsigned int *cnt;
    struct pvr2_stream *sp = sp;
    switch (bp->state) {
    case pvr2_buffer_state_idle:
        cnt = &sp->i_count;
        break;
    case pvr2_buffer_state_queued:
        cnt = &sp->q_count;
        break;
    case pvr2_buffer_state_ready:
        cnt = &sp->r_count;
        break;
    default:
        return;
    }
    list_del_init(&bp->list_overhead);
    *cnt--;
    bp->state = pvr2_buffer_state_none;
}

static int pvr2_buffer_set_none(struct pvr2_buffer *bp)
{
    int fl;
    unsigned long irq_flags;
    struct pvr2_stream *sp;
    spin_lock_irqsave(&sp->list_lock,irq_flags);
    sp = bp->stream;
    fl = (sp->r_count == 0);
    pvr2_buffer_remove(bp);
    spin_unlock_irqrestore(&sp->list_lock,irq_flags);
    return fl;
}

static int pvr2_buffer_set_ready(struct pvr2_buffer *bp)
{
    int fl;
    unsigned long irq_flags;
    struct pvr2_stream *sp;
    spin_lock_irqsave(&sp->list_lock,irq_flags);
    sp = bp->stream;
    fl = (sp->r_count == 0);
    pvr2_buffer_remove(bp);
    list_add_tail(&bp->list_overhead,&sp->ready_list);
    bp->state = pvr2_buffer_state_ready;
    (sp->r_count)++;
    spin_unlock_irqrestore(&sp->list_lock,irq_flags);
    return fl;
}

static void pvr2_buffer_set_idle(struct pvr2_buffer *bp)
{
    unsigned long irq_flags;
    struct pvr2_stream *sp;
    spin_lock_irqsave(&sp->list_lock,irq_flags);
    sp = bp->stream;
    pvr2_buffer_remove(bp);
    list_add_tail(&bp->list_overhead,&sp->idle_list);
    bp->state = pvr2_buffer_state_idle;
    (sp->i_count)++;
    spin_unlock_irqrestore(&sp->list_lock,irq_flags);
}

static void pvr2_buffer_set_queued(struct pvr2_buffer *bp)
{
    unsigned long irq_flags;
    struct pvr2_stream *sp;
    spin_lock_irqsave(&sp->list_lock,irq_flags);
    sp = bp->stream;
    pvr2_buffer_remove(bp);
    list_add_tail(&bp->list_overhead,&sp->queued_list);
    bp->state = pvr2_buffer_state_queued;
    (sp->q_count)++;
    spin_unlock_irqrestore(&sp->list_lock,irq_flags);
}

static void pvr2_buffer_wipe(struct pvr2_buffer *bp)
{
    if (bp->state == pvr2_buffer_state_queued) {
        usb_kill_urb(bp->purb);
    }
}

static int pvr2_buffer_init(struct pvr2_buffer *bp,struct pvr2_stream *sp)
{
    memset(bp,0,sizeof(*bp));
    bp->stream = sp;
    bp->state = pvr2_buffer_state_none;
    INIT_LIST_HEAD(&bp->list_overhead);
    bp->purb = usb_alloc_urb(0,GFP_KERNEL);
    if (! bp->purb) return -ENOMEM;
    return 0;
}

static void pvr2_buffer_done(struct pvr2_buffer *bp)
{
    pvr2_buffer_wipe(bp);
    pvr2_buffer_set_none(bp);
    bp->stream = 0;
    if (bp->purb) usb_free_urb(bp->purb);
}

static int pvr2_stream_buffer_count(struct pvr2_stream *sp,unsigned int cnt)
{
    int ret;
    unsigned int scnt;

    // Allocate buffers point array in multiples of 32 entries
    if (cnt == sp->buffer_total_count) return 0;
    scnt = cnt & ~0x1f;
    if (cnt > scnt) scnt += 0x20;

    if (cnt > sp->buffer_total_count) {
        if (scnt > sp->buffer_slot_count) {
            struct pvr2_buffer **nb;
            nb = kmalloc(scnt * sizeof(*nb),GFP_KERNEL);
            if (!nb) return -ENOMEM;
            if (sp->buffer_slot_count) {
                memcpy(nb,sp->buffers,sp->buffer_slot_count * sizeof(*nb));
                kfree(sp->buffers);
            }
            sp->buffers = nb;
            sp->buffer_slot_count = scnt;
        }
        while (sp->buffer_total_count < cnt) {
            struct pvr2_buffer *bp;
            bp = kmalloc(sizeof(*bp),GFP_KERNEL);
            if (!bp) return -ENOMEM;
            ret = pvr2_buffer_init(bp,sp);
            if (ret) {
                kfree(bp);
                return -ENOMEM;
            }
            bp->id = sp->buffer_total_count;
            sp->buffers[sp->buffer_total_count] = bp;
            (sp->buffer_total_count)++;
            pvr2_buffer_set_idle(bp);
        }
    } else {
        while (sp->buffer_total_count > cnt) {
            struct pvr2_buffer *bp;
            bp = sp->buffers[sp->buffer_total_count - 1];
            (sp->buffer_total_count)--;
            pvr2_buffer_done(bp);
        }
        if (scnt < sp->buffer_slot_count) {
            struct pvr2_buffer **nb = 0;
            if (scnt) {
                nb = kmalloc(scnt * sizeof(*nb),GFP_KERNEL);
                if (!nb) return -ENOMEM;
                memcpy(nb,sp->buffers,scnt * sizeof(*nb));
            }
            kfree(sp->buffers);
            sp->buffers = nb;
            sp->buffer_slot_count = scnt;
        }
    }
    return 0;
}

static int pvr2_stream_achieve_buffer_count(struct pvr2_stream *sp)
{
    struct pvr2_buffer *bp;
    unsigned int cnt;

    if (sp->buffer_total_count == sp->buffer_target_count) return 0;

    if (sp->buffer_total_count < sp->buffer_target_count) {
        return pvr2_stream_buffer_count(sp,sp->buffer_target_count);
    }

    cnt = 0;
    while (sp->buffer_total_count > sp->buffer_target_count) {
        bp = sp->buffers[sp->buffer_total_count - 1];
        if (bp->state != pvr2_buffer_state_idle) break;
        cnt++;
    }
    if (cnt) pvr2_stream_buffer_count(sp,sp->buffer_target_count - cnt);

    return 0;
}

static void pvr2_stream_internal_flush(struct pvr2_stream *sp)
{
    struct pvr2_buffer *bp1,*bp2;
    list_for_each_entry_safe(bp1,bp2,&sp->queued_list,list_overhead) {
        pvr2_buffer_wipe(bp1);
        pvr2_buffer_set_idle(bp1);
    }
    if (sp->buffer_total_count != sp->buffer_target_count) {
        pvr2_stream_achieve_buffer_count(sp);
    }
}

static void pvr2_stream_init(struct pvr2_stream *sp)
{
    memset(sp,0,sizeof(*sp));
    spin_lock_init(&sp->list_lock);
    init_MUTEX(&sp->sem);
    INIT_LIST_HEAD(&sp->queued_list);
    INIT_LIST_HEAD(&sp->ready_list);
    INIT_LIST_HEAD(&sp->idle_list);
}

static void pvr2_stream_done(struct pvr2_stream *sp)
{
    down(&sp->sem);
    pvr2_stream_internal_flush(sp);
    pvr2_stream_buffer_count(sp,0);
    up(&sp->sem);
}

static void buffer_complete(struct urb *urb, struct pt_regs *regs)
{
    struct pvr2_buffer *bp = urb->context;
    struct pvr2_stream *sp = bp->stream;
    bp->used_count = 0;
    if ((!(urb->status)) ||
        (urb->status == -ENOENT) ||
        (urb->status == -ECONNRESET) ||
        (urb->status == -ESHUTDOWN)) {
        bp->used_count = urb->actual_length;
    }
    bp->status = urb->status;
    pvr2_buffer_set_ready(bp);
    if (sp && sp->callback_func) {
        sp->callback_func(sp->callback_data);
    }
}

struct pvr2_stream *pvr2_stream_create(void)
{
    struct pvr2_stream *sp;
    sp = kmalloc(sizeof(*sp),GFP_KERNEL);
    if (sp) pvr2_stream_init(sp);
    return sp;
}

void pvr2_stream_destroy(struct pvr2_stream *sp)
{
    if (!sp) return;
    pvr2_stream_done(sp);
    kfree(sp);
}

void pvr2_stream_setup(struct pvr2_stream *sp,
                       struct usb_device *dev,
                       unsigned int pipe)
{
    down(&sp->sem);
    sp->dev = dev;
    sp->pipe = pipe;
    up(&sp->sem);
}

void pvr2_stream_set_callback(struct pvr2_stream *sp,
                              pvr2_stream_callback func,
                              void *data)
{
    unsigned long irq_flags;
    down(&sp->sem);
    spin_lock_irqsave(&sp->list_lock,irq_flags);
    sp->callback_data = data;
    sp->callback_func = func;
    spin_unlock_irqrestore(&sp->list_lock,irq_flags);
    up(&sp->sem);
}

/* Query / set the nominal buffer count */
int pvr2_stream_get_buffer_count(struct pvr2_stream *sp)
{
    return sp->buffer_target_count;
}

int pvr2_stream_set_buffer_count(struct pvr2_stream *sp,unsigned int cnt)
{
    int ret;
    if (sp->buffer_target_count == cnt) return 0;
    down(&sp->sem);
    sp->buffer_target_count = cnt;
    ret = pvr2_stream_achieve_buffer_count(sp);
    up(&sp->sem);
    return ret;
}

struct pvr2_buffer *pvr2_stream_get_idle_buffer(struct pvr2_stream *sp)
{
    struct list_head *lp = sp->idle_list.next;
    if (!lp) return 0;
    return list_entry(lp,struct pvr2_buffer,list_overhead);
}

struct pvr2_buffer *pvr2_stream_get_ready_buffer(struct pvr2_stream *sp)
{
    struct list_head *lp = sp->ready_list.next;
    if (!lp) return 0;
    return list_entry(lp,struct pvr2_buffer,list_overhead);
}

struct pvr2_buffer *pvr2_stream_get_buffer(struct pvr2_stream *sp,int id)
{
    if (id < 0) return 0;
    if (id >= sp->buffer_total_count) return 0;
    return sp->buffers[id];
}

int pvr2_stream_get_ready_count(struct pvr2_stream *sp)
{
    return sp->r_count;
}

int pvr2_stream_get_idle_count(struct pvr2_stream *sp)
{
    return sp->i_count;
}

void pvr2_stream_flush(struct pvr2_stream *sp)
{
    down(&sp->sem);
    pvr2_stream_internal_flush(sp);
    up(&sp->sem);
}

void pvr2_stream_kill(struct pvr2_stream *sp)
{
    struct pvr2_buffer *bp;
    down(&sp->sem);
    pvr2_stream_internal_flush(sp);
    while ((bp = pvr2_stream_get_ready_buffer(sp)) != 0) {
        pvr2_buffer_set_idle(bp);
    }
    if (sp->buffer_total_count != sp->buffer_target_count) {
        pvr2_stream_achieve_buffer_count(sp);
    }
    up(&sp->sem);
}

int pvr2_buffer_queue(struct pvr2_buffer *bp)
{
    struct pvr2_stream *sp;
    if (!bp) return -EINVAL;
    sp = bp->stream;
    down(&sp->sem);
    pvr2_buffer_wipe(bp);
    pvr2_buffer_set_queued(bp);
    bp->status = -EINPROGRESS;
    usb_fill_bulk_urb(bp->purb,      // struct urb *urb
                      sp->dev,       // struct usb_device *dev
                      sp->pipe,      // unsigned int pipe
                      bp->ptr,       // void *transfer_buffer
                      bp->max_count, // int buffer_length
                      buffer_complete,
                      bp);
    usb_submit_urb(bp->purb,GFP_KERNEL);
    up(&sp->sem);
    return 0;
}

int pvr2_buffer_idle(struct pvr2_buffer *bp)
{
    struct pvr2_stream *sp;
    if (!bp) return -EINVAL;
    sp = bp->stream;
    down(&sp->sem);
    pvr2_buffer_wipe(bp);
    pvr2_buffer_set_idle(bp);
    if (sp->buffer_total_count != sp->buffer_target_count) {
        pvr2_stream_achieve_buffer_count(sp);
    }
    up(&sp->sem);
    return 0;
}

int pvr2_buffer_set_buffer(struct pvr2_buffer *bp,void *ptr,unsigned int cnt)
{
    int ret = 0;
    unsigned long irq_flags;
    struct pvr2_stream *sp;
    if (!bp) return -EINVAL;
    sp = bp->stream;
    down(&sp->sem);
    spin_lock_irqsave(&sp->list_lock,irq_flags);
    if (bp->state == pvr2_buffer_state_queued) {
        ret = -EPERM;
    } else {
        bp->ptr = ptr;
        bp->max_count = cnt;
    }
    spin_unlock_irqrestore(&sp->list_lock,irq_flags);
    up(&sp->sem);
    return ret;
}

unsigned int pvr2_buffer_get_count(struct pvr2_buffer *bp)
{
    return bp->used_count;
}

int pvr2_buffer_get_status(struct pvr2_buffer *bp)
{
    return bp->status;
}

enum pvr2_buffer_state pvr2_buffer_get_state(struct pvr2_buffer *bp)
{
    return bp->state;
}

int pvr2_buffer_get_id(struct pvr2_buffer *bp)
{
    return bp->id;
}

/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
