/*
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/kernel.h>
#include <linux/videodev.h>

#include "pvrusb2-v4l2-internal.h"
#include "pvrusb2-buffer.h"
#include "pvrusb2-stream.h"
#include "pvrusb2-debug.h"

static int video_nr = -1;       // Number of Video Devices Initialized
module_param(video_nr, int, -1);
MODULE_PARM_DESC(video_nr, "Offset for device's minor");

#define V4L2_CID_PVR_SRATE          (V4L2_CID_PRIVATE_BASE)
#define V4L2_CID_PVR_AUDIOBITRATE   (V4L2_CID_PRIVATE_BASE+1)
#define V4L2_CID_PVR_AUDIOCRC       (V4L2_CID_PRIVATE_BASE+2)
#define V4L2_CID_PVR_AUDIOEMPHASIS  (V4L2_CID_PRIVATE_BASE+3)
#define V4L2_CID_PVR_VBR            (V4L2_CID_PRIVATE_BASE+4)
#define V4L2_CID_PVR_VIDEOBITRATE   (V4L2_CID_PRIVATE_BASE+5)
#define V4L2_CID_PVR_VIDEOPEAK      (V4L2_CID_PRIVATE_BASE+6)
#define V4L2_CID_PVR_VIDEOSTANDARD  (V4L2_CID_PRIVATE_BASE+7)
#define V4L2_CID_PVR_INPUT          (V4L2_CID_PRIVATE_BASE+8)
#define V4L2_CID_PVR_AUDIOMODE      (V4L2_CID_PRIVATE_BASE+9)
#define V4L2_CID_PVR_FREQUENCY      (V4L2_CID_PRIVATE_BASE+10)
#define V4L2_CID_PVR_HRES           (V4L2_CID_PRIVATE_BASE+11)
#define V4L2_CID_PVR_VRES           (V4L2_CID_PRIVATE_BASE+12)

#define V4L2_CID_PVR_MAX            (V4L2_CID_PRIVATE_BASE+12)

struct v4l2_capability pvr_capability ={
    .driver         = "pvrusb2",
    .card           = "Hauppauge WinTV pvr-usb2",
    .bus_info       = "usb",
    .version        = KERNEL_VERSION(0,6,0),
    .capabilities   = (V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_VBI_CAPTURE
                       | V4L2_CAP_TUNER | V4L2_CAP_AUDIO
                       | V4L2_CAP_READWRITE | V4L2_CAP_STREAMING),
    .reserved       = {0,0,0,0}
};

static struct v4l2_tuner pvr_v4l2_tuners[]= {
    {
        .index      = 0,
        .name       = "TV Tuner",
        .type           = V4L2_TUNER_ANALOG_TV,
        .capability     = (V4L2_TUNER_CAP_NORM | V4L2_TUNER_CAP_STEREO |
                           V4L2_TUNER_CAP_LANG1 | V4L2_TUNER_CAP_LANG2),
        .rangelow   = 0,
        .rangehigh  = 0,
        .rxsubchans     = V4L2_TUNER_SUB_STEREO,
        .audmode        = V4L2_TUNER_MODE_STEREO,
        .signal         = 0,
        .afc            = 0,
        .reserved       = {0,0,0,0}
    }
/*    {
      .index    = 1,
      .name   = "Radio Tuner",
      .type             = V4L2_TUNER_RADIO,
      .capability       = (V4L2_TUNER_CAP_STEREO),
      .rangelow   = 0,
      .rangehigh  = 0,
      .rxsubchans       = 0,
      .audmode  = V4L2_TUNER_MODE_STEREO,
      .signal           = 0,
      .afc              = 0,
      .reserved = {0,0,0,0}
      }*/
};

struct v4l2_standard pvr_standards[] = {
    [PVR2_CVAL_VIDEOSTANDARD_PAL_BG] = {
        .id             = V4L2_STD_PAL_BG,
        .frameperiod    =
        {
            .numerator  = 1,
            .denominator= 25
        },
        .framelines     = 625,
        .reserved       = {0,0,0,0}
    },
    [PVR2_CVAL_VIDEOSTANDARD_PAL_I] = {
        .id             = V4L2_STD_PAL_I,
        .frameperiod    =
        {
            .numerator  = 1,
            .denominator= 25
        },
        .framelines     = 625,
        .reserved       = {0,0,0,0}
    },
    [PVR2_CVAL_VIDEOSTANDARD_PAL_DK] = {
        .id             = V4L2_STD_PAL_DK,
        .frameperiod    =
        {
            .numerator  = 1,
            .denominator= 25
        },
        .framelines     = 625,
        .reserved       = {0,0,0,0}
    },
    [PVR2_CVAL_VIDEOSTANDARD_SECAM_L] = {
        .id             = V4L2_STD_SECAM,
        .frameperiod    =
        {
            .numerator  = 1,
            .denominator= 25
        },
        .framelines     = 625,
        .reserved       = {0,0,0,0}
    },
    [PVR2_CVAL_VIDEOSTANDARD_NTSC_M] = {
        .id             = V4L2_STD_NTSC,
        .frameperiod    =
        {
            .numerator =  1001,
            .denominator= 30000
        },
        .framelines     = 525,
        .reserved       = {0,0,0,0}
    }
};

struct v4l2_fmtdesc pvr_fmtdesc [] = {
    {
        .index          = 0,
        .type           = V4L2_BUF_TYPE_VIDEO_CAPTURE,
        .flags          = V4L2_FMT_FLAG_COMPRESSED,
        .description    = "MPEG1/2",
        // This should really be V4L2_PIX_FMT_MPEG, but xawtv breaks
        // when I do that.
        .pixelformat    = 0, // V4L2_PIX_FMT_MPEG,
        .reserved       = { 0, 0, 0, 0 }
    }
};

struct v4l2_format pvr_format [] = {
    {
        .type   = V4L2_BUF_TYPE_VIDEO_CAPTURE,
        .fmt    = {
            .pix        = {
                .width          = 720,
                .height             = 576,
                // This should really be V4L2_PIX_FMT_MPEG, but xawtv
                // breaks when I do that.
                .pixelformat    = 0, // V4L2_PIX_FMT_MPEG,
                .field          = V4L2_FIELD_INTERLACED,
                .bytesperline   = 0,  // doesn't make sense here
                .sizeimage          = (32*1024), //FIXME : Don't know what to put here...
                .colorspace     = 0, // doesn't make sense here
                .priv           = 0
            }
        }
    },{
        .type   = V4L2_BUF_TYPE_VBI_CAPTURE,
        .fmt    = {
            .vbi        = {
                .sampling_rate = 27000000,
                .offset = 248,
                .samples_per_line = 1443,
                .sample_format = V4L2_PIX_FMT_GREY,
                .start = { 0, 0 },
                .count = { 0, 0 },
                .flags = 0,
                .reserved = { 0, 0 }
            }
        }
    }
};

static int cnv_cid_v4l2_pvr2(int id)
{
    switch (id) {
    case V4L2_CID_BRIGHTNESS:
        return PVR2_CID_BRIGHTNESS;
    case V4L2_CID_SATURATION:
        return PVR2_CID_SATURATION;
    case V4L2_CID_CONTRAST:
        return PVR2_CID_CONTRAST;
    case V4L2_CID_HUE:
        return PVR2_CID_HUE;
    case V4L2_CID_AUDIO_VOLUME:
        return PVR2_CID_VOLUME;
    case V4L2_CID_AUDIO_BALANCE:
        return PVR2_CID_BALANCE;
    case V4L2_CID_AUDIO_BASS:
        return PVR2_CID_BASS;
    case V4L2_CID_AUDIO_TREBLE:
        return PVR2_CID_TREBLE;
    case V4L2_CID_AUDIO_MUTE:
        return PVR2_CID_MUTE;
    case V4L2_CID_PVR_SRATE:
        return PVR2_CID_SRATE;
    case V4L2_CID_PVR_AUDIOBITRATE:
        return PVR2_CID_AUDIOBITRATE;
    case V4L2_CID_PVR_AUDIOCRC:
        return PVR2_CID_AUDIOCRC;
    case V4L2_CID_PVR_AUDIOEMPHASIS:
        return PVR2_CID_AUDIOEMPHASIS;
    case V4L2_CID_PVR_VBR:
        return PVR2_CID_VBR;
    case V4L2_CID_PVR_VIDEOBITRATE:
        return PVR2_CID_AVERAGEVIDEOBITRATE;
    case V4L2_CID_PVR_VIDEOPEAK:
        return PVR2_CID_PEAKVIDEOBITRATE;
    case V4L2_CID_PVR_INPUT:
        return PVR2_CID_INPUT;
    case V4L2_CID_PVR_AUDIOMODE:
        return PVR2_CID_AUDIOMODE;
    case V4L2_CID_PVR_FREQUENCY:
        return PVR2_CID_FREQUENCY;
    case V4L2_CID_PVR_HRES:
        return PVR2_CID_HRES;
    case V4L2_CID_PVR_VRES:
        return PVR2_CID_VRES;
    }
    return -1;
}

#ifdef notdef
static int cnv_cid_pvr2_v4l2(int id)
{
    switch (id) {
    case PVR2_CID_BRIGHTNESS:
        return V4L2_CID_BRIGHTNESS;
    case PVR2_CID_SATURATION:
        return V4L2_CID_SATURATION;
    case PVR2_CID_CONTRAST:
        return V4L2_CID_CONTRAST;
    case PVR2_CID_HUE:
        return V4L2_CID_HUE;
    case PVR2_CID_VOLUME:
        return V4L2_CID_AUDIO_VOLUME;
    case PVR2_CID_BALANCE:
        return V4L2_CID_AUDIO_BALANCE;
    case PVR2_CID_BASS:
        return V4L2_CID_AUDIO_BASS;
    case PVR2_CID_TREBLE:
        return V4L2_CID_AUDIO_TREBLE;
    case PVR2_CID_MUTE:
        return V4L2_CID_AUDIO_MUTE;

        return id + V4L2_CID_PRIVATE_BASE;
    }
    return -1;
}
#endif

#ifdef notdef
static const char *get_cmd_code(int cmd)
{
    switch (cmd) {
    case VIDIOC_QUERYCAP:
        return "VIDIOC_QUERYCAP";
    case VIDIOC_G_PRIORITY:
        return "VIDIOC_G_PRIORITY";
    case VIDIOC_S_PRIORITY:
        return "VIDIOC_S_PRIORITY";
    case VIDIOC_ENUMSTD:
        return "VIDIOC_ENUMSTD";
    case VIDIOC_G_STD:
        return "VIDIOC_G_STD";
    case VIDIOC_S_STD:
        return "VIDIOC_S_STD";
    case VIDIOC_ENUMINPUT:
        return "VIDIOC_ENUMINPUT";
    case VIDIOC_G_INPUT:
        return "VIDIOC_G_INPUT";
    case VIDIOC_S_INPUT:
        return "VIDIOC_S_INPUT";
    case VIDIOC_ENUMAUDIO:
        return "VIDIOC_ENUMAUDIO";
    case VIDIOC_G_AUDIO:
        return "VIDIOC_G_AUDIO";
    case VIDIOC_S_AUDIO:
        return "VIDIOC_S_AUDIO";
    case VIDIOC_G_TUNER:
        return "VIDIOC_G_TUNER";
    case VIDIOC_S_TUNER:
        return "VIDIOC_S_TUNER";
    case VIDIOC_S_FREQUENCY:
        return "VIDIOC_S_FREQUENCY";
    case VIDIOC_G_FREQUENCY:
        return "VIDIOC_G_FREQUENCY";
    case VIDIOC_ENUM_FMT:
        return "VIDIOC_ENUM_FMT";
    case VIDIOC_G_FMT:
        return "VIDIOC_G_FMT";
    case VIDIOC_TRY_FMT:
        return "VIDIOC_TRY_FMT";
    case VIDIOC_S_FMT:
        return "VIDIOC_S_FMT";
    case VIDIOC_QBUF:
        return "VIDIOC_QBUF";
    case VIDIOC_QUERYBUF:
        return "VIDIOC_QUERYBUF";
    case VIDIOC_DQBUF:
        return "VIDIOC_DQBUF";
    case VIDIOC_REQBUFS:
        return "VIDIOC_REQBUFS";
    case VIDIOC_STREAMON:
        return "VIDIOC_STREAMON";
    case VIDIOC_STREAMOFF:
        return "VIDIOC_STREAMOFF";
    case VIDIOC_QUERYCTRL:
        return "VIDIOC_QUERYCTRL";
    case VIDIOC_QUERYMENU:
        return "VIDIOC_QUERYMENU";
    case VIDIOC_G_CTRL:
        return "VIDIOC_G_CTRL";
    case VIDIOC_S_CTRL:
        return "VIDIOC_S_CTRL";
    case VIDIOCGMBUF:
        return "VIDIOCGMBUF";
    default :
        break;
    }
    return "unknown";
}
#endif


/*
 * pvr_ioctl()
 *
 * This is part of Video 4 Linux API. The procedure handles ioctl() calls.
 *
 */
static int pvr2_v4l2_do_ioctl(struct inode *inode, struct file *file,
                              unsigned int cmd, void *arg)
{
    struct pvr2_v4l2_fh *fh = file->private_data;
    struct pvr2_v4l2 *vp = fh->vhead;
    struct pvr2_v4l2_stream *stream = fh->stream;
    struct pvr2_hdw *hdw = vp->channel.mc_head->hdw;
    int ret = -EINVAL;

//    info("pvr2_v4l2_do_ioctl cmd=%s (%d)",get_cmd_code(cmd),cmd);

    if (!pvr2_hdw_dev_ok(hdw)) {
        err("ioctl failed - bad or no context");
        return -EFAULT;
    }

    /* check priority */
    switch (cmd) {
    case VIDIOC_S_CTRL:
    case VIDIOC_S_STD:
    case VIDIOC_S_INPUT:
    case VIDIOC_S_TUNER:
    case VIDIOC_S_FREQUENCY:
        ret = v4l2_prio_check(&vp->prio, &fh->prio);
        if (ret)
            return ret;
    }

    switch (cmd) {
    case VIDIOC_QUERYCAP:
    {
        struct v4l2_capability *cap = arg;

        memcpy(cap, &pvr_capability, sizeof(struct v4l2_capability));

        ret = 0;
        break;
    }

    case VIDIOC_G_PRIORITY:
    {
        enum v4l2_priority *p = arg;

        *p = v4l2_prio_max(&vp->prio);
        ret = 0;
        break;
    }

    case VIDIOC_S_PRIORITY:
    {
        enum v4l2_priority *prio = arg;

        ret = v4l2_prio_change(&vp->prio, &fh->prio, *prio);
        break;
    }

    case VIDIOC_ENUMSTD:
    {

        struct v4l2_standard *vs = (struct v4l2_standard *)arg;
        int idx = vs->index;

        if ((vs->index < PVR2_CVAL_VIDEOSTANDARD_MIN) ||
            (vs->index > PVR2_CVAL_VIDEOSTANDARD_MAX)) {
            break;
        }

        memcpy(vs, &pvr_standards[idx], sizeof(struct v4l2_standard));
        vs->index = idx;
        strlcpy(vs->name,
                pvr2_hdw_get_ctl_value_name(hdw,
                                            PVR2_CID_VIDEOSTANDARD,
                                            idx),
                sizeof(vs->name));

        ret = 0;
        break;
    }

    case VIDIOC_G_STD:
    {
        v4l2_std_id *vs = (v4l2_std_id *)arg;

        switch (pvr2_hdw_get_ctl_value(hdw,PVR2_CID_VIDEOSTANDARD)) {
        default:
        case PVR2_CVAL_VIDEOSTANDARD_NTSC_M:
            *vs = V4L2_STD_NTSC;
            break;
        case PVR2_CVAL_VIDEOSTANDARD_SECAM_L:
            *vs = V4L2_STD_SECAM;
            break;
        case PVR2_CVAL_VIDEOSTANDARD_PAL_BG:
            *vs = V4L2_STD_PAL_BG;
            break;
        case PVR2_CVAL_VIDEOSTANDARD_PAL_I:
            *vs = V4L2_STD_PAL_I;
            break;
        case PVR2_CVAL_VIDEOSTANDARD_PAL_DK:
            *vs = V4L2_STD_PAL_DK;
            break;
        }
        ret = 0;
        break;
    }

    case VIDIOC_S_STD:
    {
        v4l2_std_id *vs = (v4l2_std_id *)arg;
        int val = PVR2_CVAL_VIDEOSTANDARD_NTSC_M;

        if (*vs & V4L2_STD_NTSC){
            val = PVR2_CVAL_VIDEOSTANDARD_NTSC_M;
        } else if (*vs & V4L2_STD_PAL_BG){
            val = PVR2_CVAL_VIDEOSTANDARD_PAL_BG;
        } else if (*vs & V4L2_STD_PAL_I){
            val = PVR2_CVAL_VIDEOSTANDARD_PAL_I;
        } else if (*vs & V4L2_STD_PAL_DK){
            val = PVR2_CVAL_VIDEOSTANDARD_PAL_DK;
        } else if (*vs & V4L2_STD_SECAM){
            val = PVR2_CVAL_VIDEOSTANDARD_SECAM_L;
        }

        pvr2_hdw_set_ctl_value(hdw,PVR2_CID_VIDEOSTANDARD,val);

        break;
    }

    case VIDIOC_ENUMINPUT:
    {
        struct v4l2_input *vi = (struct v4l2_input *)arg;
        struct v4l2_input tmp;

//            pvr_debug("VIDIOC_ENUM_INPUT");

        if ((vi->index > PVR2_CVAL_INPUT_MAX) ||
            (vi->index < PVR2_CVAL_INPUT_MIN)) {
            break;
        }

        memset(&tmp,0,sizeof(tmp));
        tmp.index = vi->index;
        switch (vi->index) {
        case PVR2_CVAL_INPUT_TV:
        case PVR2_CVAL_INPUT_RADIO:
            tmp.type = V4L2_INPUT_TYPE_TUNER;
            break;
        case PVR2_CVAL_INPUT_SVIDEO:
        case PVR2_CVAL_INPUT_COMPOSITE:
            tmp.type = V4L2_INPUT_TYPE_CAMERA;
            break;
        }

        strlcpy(tmp.name,pvr2_hdw_get_ctl_value_name(hdw,PVR2_CID_INPUT,
                                                     vi->index),
                sizeof(tmp.name));

        /* Don't bother with audioset, since this driver currently
           always switches the audio whenever the video is switched. */

        /* Handling std is a tougher problem.  It doesn't make sense
           in cases where a device might be multi-standard.  We could
           just copy out the current value for the standard, but it
           can change over time.  For now just leave it zero. */

        memcpy(vi, &tmp, sizeof(tmp));

        ret = 0;
        break;
    }

    case VIDIOC_G_INPUT:
    {
        struct v4l2_input *vi = (struct v4l2_input *)arg;
        vi->index = pvr2_hdw_get_ctl_value(hdw,PVR2_CID_INPUT);
        ret = 0;
        break;
    }

    case VIDIOC_S_INPUT:
    {
        struct v4l2_input *vi = (struct v4l2_input *)arg;
        ret = 0;
        if (pvr2_hdw_set_ctl_value(hdw,PVR2_CID_INPUT,vi->index)) {
            ret = -EINVAL;
        }
        break;
    }

    case VIDIOC_ENUMAUDIO:
    {
        ret = -EINVAL;
        break;
    }

    case VIDIOC_G_AUDIO:
    {
        ret = -EINVAL;
        break;
    }

    case VIDIOC_S_AUDIO:
    {
        ret = -EINVAL;
        break;
    }
    case VIDIOC_G_TUNER:
    {
        struct v4l2_tuner *vt = (struct v4l2_tuner *)arg;
        unsigned int status_mask;
        if (vt->index !=0) break;

        status_mask = pvr2_hdw_get_signal_status(hdw);

        memcpy(vt, &pvr_v4l2_tuners[vt->index], sizeof(struct v4l2_tuner));

        vt->signal = 0;
        if (status_mask & PVR2_SIGNAL_OK) {
            if (status_mask & PVR2_SIGNAL_STEREO) {
                vt->rxsubchans = V4L2_TUNER_SUB_STEREO;
            } else {
                vt->rxsubchans = V4L2_TUNER_SUB_MONO;
            }
            if (status_mask & PVR2_SIGNAL_SAP) {
                vt->rxsubchans |= (V4L2_TUNER_SUB_LANG1 |
                                   V4L2_TUNER_SUB_LANG2);
            }
            vt->signal = 65535;
        }

        switch (pvr2_hdw_get_ctl_value(hdw,PVR2_CID_AUDIOMODE)) {
        case PVR2_CVAL_AUDIOMODE_MONO:
            vt->audmode = V4L2_TUNER_MODE_MONO;
            break;
        case PVR2_CVAL_AUDIOMODE_STEREO:
            vt->audmode = V4L2_TUNER_MODE_STEREO;
            break;
        case PVR2_CVAL_AUDIOMODE_LANG1:
            vt->audmode = V4L2_TUNER_MODE_LANG1;
            break;
        case PVR2_CVAL_AUDIOMODE_LANG2:
            vt->audmode = V4L2_TUNER_MODE_LANG2;
            break;
        case PVR2_CVAL_AUDIOMODE_SAP:
            vt->audmode = V4L2_TUNER_MODE_SAP;
            break;
        }

        ret = 0;
        break;
    }

    case VIDIOC_S_TUNER:
    {
        struct v4l2_tuner *vt=(struct v4l2_tuner *)arg;
        int val = PVR2_CVAL_AUDIOMODE_STEREO;

        if (vt->index != 0)
            break;

        switch (vt->audmode) {
        case V4L2_TUNER_MODE_MONO:
            val = PVR2_CVAL_AUDIOMODE_MONO;
            break;
        case V4L2_TUNER_MODE_STEREO:
            val = PVR2_CVAL_AUDIOMODE_STEREO;
            break;
        case V4L2_TUNER_MODE_LANG1:
            val = PVR2_CVAL_AUDIOMODE_LANG1;
            break;
        case V4L2_TUNER_MODE_SAP: // Also LANG2
            val = PVR2_CVAL_AUDIOMODE_SAP;
            break;
        }

        pvr2_hdw_set_ctl_value(hdw,PVR2_CID_AUDIOMODE,val);
        ret = 0;
    }

    case VIDIOC_S_FREQUENCY:
    {
        const struct v4l2_frequency *vf = (struct v4l2_frequency *)arg;

        pvr2_hdw_set_ctl_value(hdw,PVR2_CID_FREQUENCY,
                               vf->frequency * 62500);

        ret = 0;
        break;
    }

    case VIDIOC_G_FREQUENCY:
    {
        struct v4l2_frequency *vf = (struct v4l2_frequency *)arg;
        int val;

        val = pvr2_hdw_get_ctl_value(hdw,PVR2_CID_FREQUENCY);

        val /= 62500;
        vf->frequency = val;

        ret = 0;
        break;
    }

    case VIDIOC_ENUM_FMT:
    {
        struct v4l2_fmtdesc *fd = (struct v4l2_fmtdesc *)arg;


//            pvr_debug("VIDIOC_ENUM_FMT");
        /* Only one format is supported : mpeg.*/
        if (fd->index != 0)
            break;

        memcpy(fd, pvr_fmtdesc, sizeof(struct v4l2_fmtdesc));
        ret = 0;
        break;
    }

    case VIDIOC_G_FMT:
    {
        struct v4l2_format *vf = (struct v4l2_format *)arg;

        switch(vf->type) {
        case V4L2_BUF_TYPE_VIDEO_CAPTURE:
            memcpy(vf, &pvr_format[PVR_FORMAT_PIX],
                   sizeof(struct v4l2_format));
            vf->fmt.pix.width =
                pvr2_hdw_get_ctl_value(hdw,PVR2_CID_HRES);
            if (pvr2_hdw_get_ctl_value(hdw,PVR2_CID_INTERLACE)) {
                vf->fmt.pix.width /= 2;
            }
            vf->fmt.pix.height =
                pvr2_hdw_get_ctl_value(hdw,PVR2_CID_VRES);
            ret = 0;
            break;
        case V4L2_BUF_TYPE_VBI_CAPTURE:
            // ????? Still need to figure out to do VBI correctly
            ret = -EINVAL;
            break;
        default:
            ret = -EINVAL;
            break;
        }
        break;
    }

    case VIDIOC_TRY_FMT:
    case VIDIOC_S_FMT:
    {
        struct v4l2_format *vf = (struct v4l2_format *)arg;

        ret = 0;
        switch(vf->type) {
        case V4L2_BUF_TYPE_VIDEO_CAPTURE: {
            int is_ntsc = (pvr2_hdw_get_ctl_value(hdw,
                                                  PVR2_CID_VIDEOSTANDARD) ==
                           PVR2_CVAL_VIDEOSTANDARD_NTSC_M);
            int hf = is_ntsc ? 480 : 576;
            int hh = (int) (hf / 2);
            int h = vf->fmt.pix.height;
            int w = vf->fmt.pix.width;

            memcpy(vf, &pvr_format[PVR_FORMAT_PIX],
                   sizeof(struct v4l2_format));
            if (w > 720)
                vf->fmt.pix.width = 720;
            vf->fmt.pix.width &= 0xff0;
            vf->fmt.pix.height = (h > hh) ? hf : hh;

            if (cmd == VIDIOC_S_FMT){
                pvr2_hdw_set_ctl_value(hdw,PVR2_CID_HRES,
                                       vf->fmt.pix.width);
                pvr2_hdw_set_ctl_value(hdw,PVR2_CID_VRES,
                                       vf->fmt.pix.height);
                pvr2_hdw_set_ctl_value(hdw,PVR2_CID_INTERLACE,
                                       (vf->fmt.pix.height != hf));
            }
        } break;
        case V4L2_BUF_TYPE_VBI_CAPTURE:
            // ????? Still need to figure out to do VBI correctly
            ret = -EINVAL;
            break;
        default:
            ret = -EINVAL;
            break;
        }
        break;
    }

    case VIDIOC_QBUF:
    {
        struct v4l2_buffer *vb = (struct v4l2_buffer *) arg;

        ret = pvr_stream_qbuf(stream, vb);
        break;
    }
    case VIDIOC_QUERYBUF:
    {
        struct v4l2_buffer *vb = (struct v4l2_buffer *) arg;

        ret = pvr_stream_querybuf(stream, vb);
        break;
    }
    case VIDIOC_DQBUF:
    {
        struct v4l2_buffer *vb = (struct v4l2_buffer *) arg;

        ret = pvr_stream_dqbuf(stream, vb, file->f_flags & O_NONBLOCK);
        break;
    }
    case VIDIOC_REQBUFS:
    {
        struct v4l2_requestbuffers *vr = (struct v4l2_requestbuffers *) arg;

        ret = pvr_stream_reqbufs(stream, fh, vr);
        break;
    }

    case VIDIOC_STREAMON:
    {
        ret = pvr_stream_streamon(stream, fh);
        break;
    }

    case VIDIOC_STREAMOFF:
    {
        ret = pvr_stream_streamoff(stream, fh);
        break;
    }

    case VIDIOC_QUERYCTRL:
    {
        struct v4l2_queryctrl *vc = (struct v4l2_queryctrl *)arg;
        int pvr2_id = cnv_cid_v4l2_pvr2(vc->id);
        if (pvr2_id < 0) {
            ret = -EINVAL;
            break;
        }

        if (pvr2_hdw_get_ctl_value_name(hdw,pvr2_id,0)) {
            vc->type = V4L2_CTRL_TYPE_MENU;
        } else {
            vc->type = V4L2_CTRL_TYPE_INTEGER;
        }
        strlcpy(vc->name,pvr2_hdw_get_ctl_name(hdw,pvr2_id),sizeof(vc->name));
        vc->minimum = pvr2_hdw_get_ctl_min_value(hdw,pvr2_id);
        vc->maximum = pvr2_hdw_get_ctl_max_value(hdw,pvr2_id);
        vc->step = 1;
        ret = 0;
        break;
    }

    case VIDIOC_QUERYMENU:
    {
        struct v4l2_querymenu *vm = (struct v4l2_querymenu *)arg;
        int pvr2_id = cnv_cid_v4l2_pvr2(vm->id);
        const char *value_name;
        if (pvr2_id < 0) {
            ret = -EINVAL;
            break;
        }

        value_name = pvr2_hdw_get_ctl_value_name(hdw,pvr2_id,vm->index);
        if (value_name) {
            strlcpy(vm->name,value_name,sizeof(vm->name));
            ret = 0;
        } else {
            ret = -EINVAL;
        }

        break;
    }

    case VIDIOC_G_CTRL:
    {
        struct v4l2_control *vc = (struct v4l2_control *)arg;
        int pvr2_id;

        pvr2_id = cnv_cid_v4l2_pvr2(vc->id);
        if (pvr2_id < 0) {
            ret = -EINVAL;
            break;
        }
        ret = 0;
        vc->value = pvr2_hdw_get_ctl_value(hdw,pvr2_id);
        break;
    }

    case VIDIOC_S_CTRL:
    {
        struct v4l2_control *vc = (struct v4l2_control *)arg;
        int pvr2_id;

        pvr2_id = cnv_cid_v4l2_pvr2(vc->id);
        if (pvr2_id < 0) {
            ret = -EINVAL;
            break;
        }

        ret = pvr2_hdw_set_ctl_value(hdw,pvr2_id,vc->value);
        break;
    }

    /* v4l1 compatibility */
    case VIDIOCGMBUF:
    {
        struct video_mbuf *vmb = (struct video_mbuf *) arg;

        ret = pvr_stream_cgmbuf(stream, fh ,vmb);
        break;
    }

    default :
        ret = v4l_compat_translate_ioctl(inode, file, cmd, arg, pvr2_v4l2_do_ioctl);
    }

    pvr2_hdw_commit_ctl(hdw);

#ifdef notdef
    if (ret < 0) {
        err("pvr2_v4l2_do_ioctl failure, ret=%d (0x%x)",ret,ret);
    } else {
        info("pvr2_v4l2_do_ioctl complete, ret=%d (0x%x)",ret,ret);
    }
#endif
    return ret;
}


static void pvr2_v4l2_destroy_no_lock(struct pvr2_v4l2 *vp)
{
    unsigned int i;

    for (i = 0 ; i < vp->stream_count ; i++) {
        if (!vp->stream[i]) continue;
        pvr_stream_free(vp->stream[i]);
        vp->stream[i] = 0;
    }

    pvr2_trace(PVR2_TRACE_STRUCT,"Destroying pvr2_v4l2 id=0x%x",
               (unsigned int)vp);
    pvr2_channel_done(&vp->channel);
    kfree(vp);
}


void pvr2_v4l2_internal_check(struct pvr2_channel *chp)
{
    struct pvr2_v4l2 *vp;
    vp = container_of(chp,struct pvr2_v4l2,channel);
    if (!vp->channel.mc_head->disconnect_flag) return;
    if (vp->vfirst) return;
    pvr2_v4l2_destroy_no_lock(vp);
}


struct pvr2_v4l2 *pvr2_v4l2_create(struct pvr2_main *mnp)
{
    unsigned int i;
    struct pvr2_v4l2 *vp;
    int fail_flag = 0;

    vp = kmalloc(sizeof(*vp),GFP_KERNEL);
    if (!vp) return vp;
    memset(vp,0,sizeof(*vp));
    pvr2_main_enter(mnp); do {
        pvr2_channel_init(&vp->channel,mnp);
        pvr2_trace(PVR2_TRACE_STRUCT,"Creating pvr2_v4l2 id=0x%x",
                   (unsigned int)vp);

        vp->channel.check_func = pvr2_v4l2_internal_check;

        /* register streams */

        vp->stream[0] = pvr_stream_alloc(vp, PVR_VID_ENDPOINT,
                                         VFL_TYPE_GRABBER,video_nr,
                                         pvr2_config_mpeg);
#ifdef notdef
        vp->stream[1] = pvr_stream_alloc(vp, PVR_VBI_ENDPOINT,
                                         VFL_TYPE_VBI, video_nr,
                                         pvr2_config_vbi);
        vp->stream[2] = pvr_stream_alloc(vp, PVR_VID_ENDPOINT,
                                         VFL_TYPE_RADIO, video_nr,
                                         pvr2_config_radio);
#endif
        vp->stream_count = 1;

    } while (0); pvr2_main_exit(mnp);

    for (i = 0; i < vp->stream_count ; i++) {
        if (vp->stream[i] != NULL) continue;
        fail_flag = !0;
        break;
    }

    if (fail_flag) {
        pvr2_main_enter(mnp); do {
            pvr2_v4l2_destroy_no_lock(vp);
        } while (0); pvr2_main_exit(mnp);
        vp = 0;
    }

    return vp;
}


int pvr2_v4l2_ioctl(struct inode *inode, struct file *file,
                    unsigned int cmd, unsigned long arg)
{

/* Temporary hack : use ivtv api until a v4l2 one is available. */
#define IVTV_IOC_G_CODEC        0xFFEE7703
#define IVTV_IOC_S_CODEC        0xFFEE7704
    if (cmd == IVTV_IOC_G_CODEC || cmd == IVTV_IOC_S_CODEC) return 0;
    return video_usercopy(inode, file, cmd, arg, pvr2_v4l2_do_ioctl);
}


int pvr2_v4l2_release(struct inode *inode, struct file *file)
{
    struct pvr2_v4l2_fh *fhp = file->private_data;
    struct pvr2_v4l2 *vp = fhp->vhead;
    struct pvr2_v4l2_stream *stream = fhp->stream;
    struct pvr2_main *mp = fhp->vhead->channel.mc_head;
    int ret = 0;

    pvr2_trace(PVR2_TRACE_OPEN_CLOSE,"pvr2_v4l2_release");

    down(&stream->stream_sem); do {
        stream->user--;
        if (stream->io_owner == fhp)
            ret = pvr_io_release(stream, fhp);
        v4l2_prio_close(&vp->prio, &fhp->prio);
        file->private_data = NULL;

        pvr2_main_enter(mp); do {
            if (fhp->vnext) {
                fhp->vnext->vprev = fhp->vprev;
            } else {
                vp->vlast = fhp->vprev;
            }
            if (fhp->vprev) {
                fhp->vprev->vnext = fhp->vnext;
            } else {
                vp->vfirst = fhp->vnext;
            }
            fhp->vnext = 0;
            fhp->vprev = 0;
            fhp->vhead = 0;
            pvr2_trace(PVR2_TRACE_STRUCT,"Destroying pvr_v4l2_fh id=0x%x",
                       (unsigned int)fhp);
            kfree(fhp);
            if (vp->channel.mc_head->disconnect_flag && !vp->vfirst) {
                pvr2_v4l2_destroy_no_lock(vp);
            }
        } while (0); pvr2_main_exit(mp);
    } while (0); up(&stream->stream_sem);
    PVR_BUG_ON(ret);
    PVR_BUG_ON(stream->mem_state == _mapped);
    return 0;
}


int pvr2_v4l2_open(struct inode *inode, struct file *file)
{
    struct video_device *vdev = video_devdata(file);
    struct pvr2_v4l2_stream *stream = (struct pvr2_v4l2_stream *)video_get_drvdata(vdev);
    struct pvr2_v4l2_fh *fhp;
    struct pvr2_v4l2 *vp = stream->v4lp;
    struct pvr2_hdw *hdw = vp->channel.hdw;
    int ret = 0;

    pvr2_trace(PVR2_TRACE_OPEN_CLOSE,"pvr2_v4l2_open");

    down(&stream->stream_sem); do {
        if (!pvr2_hdw_dev_ok(hdw)) {
            pvr_debug("pvr2_v4l2_open: hardware not ready");
            ret = -EIO;
            break;
        }

        fhp = kmalloc(sizeof(*fhp),GFP_KERNEL);
        if (!fhp) {
            ret = -ENOMEM;
            break;
        }
        memset(fhp,0,sizeof(*fhp));

        pvr2_main_enter(vp->channel.mc_head); do {
            pvr2_trace(PVR2_TRACE_STRUCT,"Creating pvr_v4l2_fh id=0x%x",
                       (unsigned int)fhp);
            fhp->vnext = 0;
            fhp->vprev = vp->vlast;
            if (vp->vlast) {
                vp->vlast->vnext = fhp;
            } else {
                vp->vfirst = fhp;
            }
            vp->vlast = fhp;
            fhp->vhead = vp;
        } while (0); pvr2_main_exit(vp->channel.mc_head);

        fhp->stream = stream;
        fhp->file = file;
        file->private_data = fhp;
        v4l2_prio_open(&vp->prio,&fhp->prio);
        if (!stream->user) pvr2_hdw_set_streaming(fhp->vhead->channel.hdw,0);
        stream->user++;

    } while (0); up(&stream->stream_sem);

    return ret;
}


/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
